#include <winsock2.h>
#include <windows.h>
#include <assert.h>
#include <math.h>
#include <stdio.h>
#include <vector>
#include <deque>

#include "event.h"
#include "thread.h"
#include "crc32.h"

#include "etherradio.h"

//#pragma comment(lib, "winmm")
//#pragma comment(lib, "ws2_32")




class WaveData {
public:
	static const int SAMPLINGRATE = 32000;
	static const int BUFFERSIZE = 1024;
private:
    struct Core {
        unsigned int reference_count;
        unsigned long buffer[BUFFERSIZE];
		WAVEHDR header;
        Core()
        {
            reference_count = 0;
			memset(&header, 0, sizeof(WAVEHDR));
			header.dwBufferLength = sizeof(unsigned long) * BUFFERSIZE;
			header.lpData = (char *)buffer;
			header.dwLoops = 1;
			reset();
        }
        ~Core()
		{
		}
		void reset()
		{
			header.dwFlags = WHDR_BEGINLOOP | WHDR_ENDLOOP;
		}
    };
    Core *core;
    void assign(Core *p)
    {
        if (p) {
            p->reference_count++;
        }
        if (core) {
            if (core->reference_count > 1) {
                core->reference_count--;
            } else {
                delete core;
            }
        }
        core = p;
    }
public:
    WaveData()
        : core(0)
    {
        assign(new Core());
    }
    WaveData(WaveData const &r)
        : core(0)
    {
        assign(r.core);
    }
    virtual ~WaveData()
    {
        assign(0);
    }
    void operator = (WaveData const &r)
    {
        assign(r.core);
    }
	WAVEHDR *header()
	{
		return &core->header;
	}
    unsigned long *buffer()
    {
		return (unsigned long *)header()->lpData;
    }
	void reset()
	{
		core->reset();
	}
};

class WavePlayerThread : public Thread {
private:
	CRITICAL_SECTION cs;
	DWORD thread_id;
	std::vector<unsigned long> incoming_buffer;
	HWAVEOUT hWaveOut;
	WAVEFORMATEX wfe;
	WaveData new_data;
	static const int buffers = 16;
	WaveData buffer[buffers];
	int front;
	int count;
	enum {
		_STOPPED,
		_BUFFERING,
		_PLAYING,
	} state;
	int offset;
	bool playing;
	bool wavewrite()
	{
		if (count > 0) {
			MMRESULT r;
			r = waveOutPrepareHeader(hWaveOut, buffer[front].header(), sizeof(WAVEHDR));
			if (r != MMSYSERR_NOERROR) {
//				_asm nop
			}
			r = waveOutWrite(hWaveOut, buffer[front].header(), sizeof(WAVEHDR));
			if (r != MMSYSERR_NOERROR) {
//				_asm nop
			}
			front = (front + 1) % buffers;
			count--;
			state = _PLAYING;
			return true;
		}
		return false;
	}
	bool addwave(unsigned long const *ptr, int len)
	{
		if (state == _STOPPED) {
			return false;
		}
		while (len > 0 && count < buffers) {
			int n = WaveData::BUFFERSIZE - offset;
			if (n > len) {
				n = len;
			}
			int index = (front + count) % buffers;
			if (n > 0) {
				unsigned long *p = buffer[index].buffer();
				for (int i = 0; i < n; i++) {
					p[offset + i] = ptr[i];
				}
				offset += n;
				ptr += n;
				len -= n;
			}
			if (offset == WaveData::BUFFERSIZE) {
				offset = 0;
				count++;
				if (state == _BUFFERING && count < 8) {
					continue;
				}
				while (wavewrite());
			}
		}
		if (state == _BUFFERING && count == 0) {
			return true; // underflow;
		}
		return false;
	}
public:
	WavePlayerThread()
	{
		InitializeCriticalSection(&cs);
		thread_id = 0;
		front = 0;
		count = 0;
		state = _STOPPED;
		offset = 0;
		playing = false;
	}
	~WavePlayerThread()
	{
		DeleteCriticalSection(&cs);
	}
private:
	void Open()
	{
		MMRESULT r;

		memset(&wfe, 0, sizeof(wfe));
		wfe.wFormatTag = WAVE_FORMAT_PCM;
		wfe.nChannels = 2;
		wfe.wBitsPerSample = 16;	//量子化ビット数
		wfe.nBlockAlign = wfe.nChannels * wfe.wBitsPerSample / 8;
		wfe.nSamplesPerSec = WaveData::SAMPLINGRATE;	//標本化周波数
		wfe.nAvgBytesPerSec = WaveData::SAMPLINGRATE * wfe.nBlockAlign;

		r = waveOutOpen(&hWaveOut, WAVE_MAPPER, &wfe, (DWORD_PTR)GetCurrentThreadId(), (DWORD_PTR)this, CALLBACK_THREAD);

		state = _STOPPED;
	}
	void Start()
	{
		state = _BUFFERING;
	}
	void Stop()
	{
		waveOutReset(hWaveOut);
		for (int i = 0; i < buffers; i++) {
			waveOutUnprepareHeader(hWaveOut, buffer[i].header(), sizeof(WAVEHDR));
			buffer[i].reset();
		}
		state = _STOPPED;
	}
	void Close()
	{
		waveOutClose(hWaveOut);
	}

	void wom_done(WAVEHDR *header)
	{
		int prep = 0;
		for (int i = 0; i < buffers; i++) {
			if (buffer[i].header()->dwFlags & WHDR_DONE) {
				waveOutUnprepareHeader(hWaveOut, buffer[i].header(), sizeof(WAVEHDR));
				buffer[i].reset();
			}
			if (buffer[i].header()->dwFlags & WHDR_PREPARED) {
				prep++;
			}
		}
		if (prep == 0) {
			state = _BUFFERING;
		}
	}
public:
	virtual void run()
	{
		thread_id = GetCurrentThreadId();

		Open();
		Start();

		while (1) {
			MSG msg;
			if (GetMessage(&msg, 0, 0, 0) == 0) {
				break;
			}
			switch (msg.message) {
			case MM_WOM_DONE:
				wom_done((WAVEHDR *)msg.lParam);
				break;
			case WM_NULL:
				EnterCriticalSection(&cs);
				if (!incoming_buffer.empty()) {
					if (playing) {
						if (addwave(&incoming_buffer[0], incoming_buffer.size())) {
							addwave(&incoming_buffer[0], incoming_buffer.size());
						}
					}
					incoming_buffer.clear();
				}
				LeaveCriticalSection(&cs);
				break;
			}
		}

		Stop();
		Close();
	}

	void add(unsigned long const *ptr, int len)
	{
		EnterCriticalSection(&cs);
		incoming_buffer.insert(incoming_buffer.end(), ptr, ptr + len);
		LeaveCriticalSection(&cs);
		PostThreadMessage(thread_id, WM_NULL, 0, 0);
	}

	void stop()
	{
		PostThreadMessage(thread_id, WM_QUIT, 0, 0);
	}

	void play(bool f)
	{
		playing = f;
	}
};

class RequestSenderThread : public Thread {
private:
	SOCKET sock;
	struct sockaddr_in addr;
	Event event;
	bool playing;
public:
	void setup(SOCKET s)
	{
		playing = false;
		sock = s;
		bind("192.168.0.100", 2000);
	}
	void bind(char const *a, int port)
	{
		addr.sin_family = AF_INET;
		addr.sin_port = htons(port);
#if 0
		addr.sin_addr.S_un.S_addr = inet_addr(a);
#else
		struct hostent *e = gethostbyname(a);
		if (e->h_addrtype == AF_INET && e->h_length > 0) {
			in_addr *p = (in_addr *)e->h_addr_list[0];
			addr.sin_addr = *p;
		}
#endif
		::bind(sock, (struct sockaddr *)&addr, sizeof(addr));
	}
	void play(bool f)
	{
		playing = f;
	}
	void send(char const *str)
	{
		sendto(sock, str, strlen(str), 0, (struct sockaddr *)&addr, sizeof(struct sockaddr_in));
	}
	virtual void run()
	{
		while (!interrupted()) {
			if (playing) {
                send("wave");
			}
			event.wait(500);
		}
	}
	virtual void stop()
	{
		Thread::stop();
		event.signal();
	}
};

class StreamReceiverThread : public Thread {
private:
	SOCKET sock;
	WavePlayerThread *playthread;
	Event event;
public:
	void setup(SOCKET s, WavePlayerThread *t)
	{
		sock = s;
		playthread = t;
	}
	virtual void run()
	{
		while (!interrupted()) {
			struct sockaddr_in from;
			from.sin_family = AF_INET;
			int srclen = 256;
			int dstlen = 256;
			std::vector<unsigned long> tmp1(srclen + 1);
			std::vector<unsigned long> tmp2(dstlen);
			int len = sizeof(sockaddr_in);
			int n = recvfrom(sock, (char *)&tmp1[0], (srclen + 1) * sizeof(unsigned long), 0, (struct sockaddr *)&from, &len);
			if (n == 1028) {
				if (crc32(0, (unsigned char const *)&tmp1[0], srclen * sizeof(unsigned long)) == ntohl(tmp1[srclen])) {
					for (int i = 0; i < dstlen; i++) {
						unsigned short *src = (unsigned short *)&tmp1[i * srclen / dstlen];
						unsigned short *dst = (unsigned short *)&tmp2[i];
						dst[0] = ntohs(src[0]) - 32768;
						dst[1] = ntohs(src[1]) - 32768;
					}
					playthread->add(&tmp2[0], dstlen);
				}
			} else {
//				_asm nop
			}
		}
	}
};

struct EtherRadio::Privates {
	SOCKET sock;
	WavePlayerThread playthread;
	RequestSenderThread sendthread;
	StreamReceiverThread recvthread;
	Privates()
	{
	}
};

EtherRadio::EtherRadio()
{
	privates = new Privates();
	privates->sock = INVALID_SOCKET;
}

EtherRadio::~EtherRadio()
{
	delete privates;
}

void EtherRadio::bind(char const *addr, int port)
{
	privates->sendthread.bind(addr, port);
}

void EtherRadio::start()
{
	if (privates->sock != INVALID_SOCKET) {
		shutdown(privates->sock, SD_BOTH);
		closesocket(privates->sock);
	}

	privates->sock = socket(AF_INET, SOCK_DGRAM, 0);

	privates->sendthread.setup(privates->sock);
	privates->recvthread.setup(privates->sock, &privates->playthread);

	privates->playthread.start();
	privates->sendthread.start();
	privates->recvthread.start();
}

void EtherRadio::stop()
{
	privates->playthread.stop();
	privates->sendthread.stop();
	privates->recvthread.stop();
	::shutdown(privates->sock, SD_BOTH);
	::closesocket(privates->sock);
	privates->playthread.join();
	privates->sendthread.join();
	privates->recvthread.join();
	privates->sock = INVALID_SOCKET;
}

void EtherRadio::play(bool f)
{
	privates->sendthread.play(f);
	privates->playthread.play(f);
}

void EtherRadio::tune_am(int f)
{
	char tmp[100];
	sprintf(tmp, "am%u", f);
	privates->sendthread.send(tmp);
}

void EtherRadio::tune_fm(int f)
{
	char tmp[100];
	sprintf(tmp, "fm%u", f);
	privates->sendthread.send(tmp);
}

void EtherRadio::mute()
{
	privates->sendthread.send("mute");
}

#if 0

int main(int argc, char **argv)
{
	WSAData wsaData;
	WSAStartup(MAKEWORD(2, 0), &wsaData);

	EtherRadio radio;

	radio.start();

	getchar();

	radio.stop();

	return 0;
}

#endif
